//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include "mgtn.h"

/**
 *
 *  Allocates a new image 
 *
 */

struct image *mgtn_image_new()
{
  struct image *image;

  image = (struct image *) xmalloc(sizeof(struct image));

  image->nrow = image->ncol = 0;
  image->gray = NULL;

  return (image);
}

/**
 *
 *  Allocates the gray level array  
 *
 */

struct image *mgtn_image_alloc(
     struct image *image,
     int nrow,
     int ncol)
{
  int size;
  struct image *tmp;
  
  if (image == NULL)
    tmp = mgtn_image_new(); 
  else
  {
    tmp = image;

    if (tmp->gray != NULL)
      error("Attempting to alloc an image which is already allocated\n");
  }

  size = nrow * ncol * sizeof(double);
  if (size <= 0)
    error("Cannot allocate an image with zero or negative size\n");

  tmp->nrow = nrow;
  tmp->ncol = ncol;
  tmp->gray = (double *) xmalloc(size);

  return tmp;
}

/**
 *
 *  Deallocates an image 
 *
 */

void mgtn_image_delete(
    struct image *image)

{
  if (image == NULL)
    error("Cannot delete image: structure is NULL\n");

  if (image->gray != NULL) 
  {
    free(image->gray);
    image->gray = NULL;
  }
  else
    error("image->gray is NULL! cannot delete\n");

  free(image);
}

/**
 *
 * Copy an image from in to out 
 *
 */

void mgtn_image_copy(
    struct image *in, 
    struct image *out)
{
  if ((!in) || (!out) || (!in->gray) || (!out->gray) 
      || (in->ncol != out->ncol) || (in->nrow != out->nrow)) 
    error("Images cannot be copied: they are NULL or images are of different sizes\n");

  memcpy(out->gray, in->gray, sizeof(double) * in->ncol*in->nrow);
}

/**
 *
 * Initialize an image to a constant value 
 *
 */

void mgtn_image_clear(
  struct image *im,
  double value)
{
  double *p;
  int i, size;

  if ((!im) || (!im->gray))
    error("Image is not allocated.\n");

  size = im->nrow * im->ncol;
  p    = im->gray;

  for(i = 0; i < size; i++, p++)
    *p = value;
}
