//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include <ctype.h>
#include "mgtn.h"

#define FALSE 0
#define TRUE  1

/**
 *
 * Read and image from disc. Returns allocated structure. 
 *
 */

struct image *mgtn_pgm_read_image(char * name)
{
  FILE *f;
  int c,items,bin,x,y;
  int xsize,ysize,depth;
  struct image *image;

  bin = FALSE;

  /* open file */
  f = fopen(name,"r");
  if(f == NULL) 
    error("Can't open input file.");

  /* read header */
  if(getc(f) != 'P') 
    error("Input file is not a PGM file!\n");

  c = getc(f);
  if (c == '2')	
    bin = FALSE;
  else if( c == '5' ) 
    bin = TRUE;
  else 
    error("This PGM file is not supported (requires P2 or P5 header)\n");

  c = getc(f);
  while (isspace(c)) c = getc(f);     /* skip to end of line*/

  while (c == '#')                    /* skip comment lines */
  {
    c = getc(f);
    while (c != '\n') c = getc(f);    /* skip to end of comment line */
    c = getc(f);
  }

  fseek(f, -1, SEEK_CUR);             /* backup one character*/

  items = fscanf(f,"%d",&xsize);
  if (items == 0) 
    error("ERROR reading PGM file. Arrived to end before expected.\n");

  items = fscanf(f,"%d",&ysize);
  if (items == 0) 
    error("ERROR reading PGM file. Arrived to end before expected.\n");

  items = fscanf(f,"%d",&depth);
  if (items == 0) 
    error("ERROR reading PGM file. Arrived to end before expected.\n");


  c = fgetc(f); /*skip the end of line*/

  /* get memory */
  image = mgtn_image_alloc(NULL, ysize, xsize);

  /* read data */
  if (bin == FALSE) {
    for(y=0;y<ysize;y++)
      for(x=0;x<xsize;x++)
      {
	items = fscanf(f, "%d", &c);
	if (items == 0) 
	  error("ERROR reading PGM file. Arrived to end before expected.\n");

	image->gray[ x + y * xsize ] = c;
      }
  }
  else
  {
    for(y=0;y<ysize;y++)
      for(x=0;x<xsize;x++)
      {
	c = fgetc(f);
	if (feof(f)) 
	  error("ERROR reading PGM file. Arrived to end before expected.\n");
	image->gray[ x + y * xsize ] = c;
      }
  }

  /* close file */
  fclose(f);

  return image;
}

/**
 *
 * Write an image to disc. 
 *
 */

void mgtn_pgm_write_image(struct image *image, char * name)
{
  FILE * f;
  double data;
  int x,y,n;

  /* open file */
  f = fopen(name,"w");
  if( f == NULL ) 
    error("Can't open output file.");

  /* write headder */
  fprintf(f,"P2\n");
  fprintf(f,"%d %d\n",image->ncol,image->nrow);
  fprintf(f,"255\n");

  /* write data */
  for(n=1,y=0; y<image->nrow; y++)
    for(x=0; x<image->ncol; x++,n++)
    {
      data = image->gray[ x + y * image->ncol ];
      if (data < 0) data = 0;
      if (data > 255 ) data = 255;
      fprintf(f,"%d ",(int)data);
      if(n==16)
      {
	fprintf(f,"\n");
	n = 0;
      }
    }

  /* close file */
  fclose(f);
}

