//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include "mgtn.h"

/**
 *
 *  Function to ensure fread reads the data that has been asked for
 *
 */

void xfread(void *ptr, size_t size, size_t nitems, FILE *stream)
{
   size_t elems;

   elems = fread(ptr, size, nitems, stream);
   
   if (elems != nitems)
    error("xfread: could not read all the data from disk.\n");
}

/**
 *  
 *  Function to ensure fwrite writes the data that has been asked for
 *
 */

void xfwrite(void *ptr, size_t size, size_t nitems, FILE *stream)
{
    size_t elems;

    elems = fwrite(ptr, size, nitems, stream);

    if (elems != nitems)
      error("xfwrite: could not write all the data to disc.\n");
}

/**
 *
 *  Function to ensure fgets reads the data that has been asked for
 *
 */

void xfgets(char *s, int n, FILE *stream)
{
  char *dummy;

  dummy = fgets(s, n, stream);

  if (dummy == NULL)
    error("xfgets: an error ocurred while reading file.\n");
}


/**
 *
 *  Reads gradient images from disc 
 *
 */

#ifdef LOW_MEMORY_COMPUTER

int mgtn_read_gradient_images(
  struct resources *DD)
{
  FILE *fp;
  unsigned int magic;
  int nrow, ncol, val, nrow_interpolated, ncol_interpolated, res;
  char fname[200];

  res = DD->curr_level;
  sprintf(fname, "gradient_%d.data", res);

  fp = fopen(fname, "r");

  if (!fp) 
    error("Could not open temporal gradient image file.\n");

  /* Read magic key */

  xfread(&magic, sizeof(unsigned int), 1, fp);

  if (magic != MAGIC_KEY)
    error("Invalid magic key on disc.\n");

  /* Read images */

  xfread(&nrow, sizeof(int), 1, fp);
  xfread(&ncol, sizeof(int), 1, fp);

  xfread(&nrow_interpolated, sizeof(int), 1, fp);
  xfread(&ncol_interpolated, sizeof(int), 1, fp);

  xfread(&val, sizeof(int), 1, fp);

  DD->grad_t0_x[res] = mgtn_image_alloc(NULL, nrow, ncol);
  DD->grad_t0_y[res] = mgtn_image_alloc(NULL, nrow, ncol);

  DD->grad_t1_x[res] = mgtn_image_alloc(NULL, nrow_interpolated, ncol_interpolated);
  DD->grad_t1_y[res] = mgtn_image_alloc(NULL, nrow_interpolated, ncol_interpolated);

  if (val)
  {
    DD->grad_t1_xx[res] = mgtn_image_alloc(NULL, nrow_interpolated, ncol_interpolated);
    DD->grad_t1_yy[res] = mgtn_image_alloc(NULL, nrow_interpolated, ncol_interpolated);
    DD->grad_t1_xy[res] = mgtn_image_alloc(NULL, nrow_interpolated, ncol_interpolated);
  }

  /* Read image data */

  xfread(DD->grad_t0_x[res]->gray, sizeof(double), nrow * ncol, fp);  
  xfread(DD->grad_t0_y[res]->gray, sizeof(double), nrow * ncol, fp);
  xfread(DD->grad_t1_x[res]->gray, sizeof(double), nrow_interpolated * ncol_interpolated, fp);
  xfread(DD->grad_t1_y[res]->gray, sizeof(double), nrow_interpolated * ncol_interpolated, fp);

  if (val)
  {
    xfread(DD->grad_t1_xx[res]->gray, sizeof(double), nrow_interpolated * ncol_interpolated, fp);
    xfread(DD->grad_t1_yy[res]->gray, sizeof(double), nrow_interpolated * ncol_interpolated, fp);
    xfread(DD->grad_t1_xy[res]->gray, sizeof(double), nrow_interpolated * ncol_interpolated, fp);
  }

  fclose(fp);

  return 1;
}

#endif

/**
 *
 *  Write gradient images to disc 
 *
 */

#ifdef LOW_MEMORY_COMPUTER

void mgtn_write_gradient_images(
  struct resources *DD)
{
  FILE *fp;
  int val, magic, nrow, ncol, nrow_interpolated, ncol_interpolated, res;
  char fname[200];

  res = DD->curr_level;
  sprintf(fname, "gradient_%d.data", res);

  fp = fopen(fname, "w");

  magic = MAGIC_KEY;

  xfwrite(&magic, sizeof(int), 1, fp);

  nrow = GET_CURRENT_NROW(DD);
  ncol = GET_CURRENT_NCOL(DD);

  xfwrite(&nrow, sizeof(int), 1, fp);
  xfwrite(&ncol, sizeof(int), 1, fp);

  ncol_interpolated = DD->grad_t1_x[res]->ncol;
  nrow_interpolated = DD->grad_t1_x[res]->nrow;

  xfwrite(&nrow_interpolated, sizeof(int), 1, fp);
  xfwrite(&ncol_interpolated, sizeof(int), 1, fp);

  val = (DD->grad_t1_xx[res] != NULL);
  xfwrite(&val, sizeof(int), 1, fp);

  printMsg("Writing gradients to disc...\n");

  xfwrite(DD->grad_t0_x[res]->gray, sizeof(double), nrow * ncol, fp);  
  xfwrite(DD->grad_t0_y[res]->gray, sizeof(double), nrow * ncol, fp);
  xfwrite(DD->grad_t1_x[res]->gray, sizeof(double), nrow_interpolated * ncol_interpolated, fp);
  xfwrite(DD->grad_t1_y[res]->gray, sizeof(double), nrow_interpolated * ncol_interpolated, fp);

  if (val)
  {
    xfwrite(DD->grad_t1_xx[res]->gray, sizeof(double), nrow_interpolated * ncol_interpolated, fp);
    xfwrite(DD->grad_t1_yy[res]->gray, sizeof(double), nrow_interpolated * ncol_interpolated, fp);
    xfwrite(DD->grad_t1_xy[res]->gray, sizeof(double), nrow_interpolated * ncol_interpolated, fp);
  }

  fclose(fp);
}

#endif

/**
 *
 * Open configuration file. If NULL pointer is passed, a 
 * temporal file is created
 *
 */

FILE *mgtn_open_configuration_file(
   char *fname)
{
  FILE *fp;

  if (fname)
  {
  fp = fopen(fname, "r");

  if (!fp) 
    error("Could not open configuration file.\n");
  }
  else
  {
    fp = tmpfile();

    if (!fp)
      error("Could not create temporary file.\n");
  }

  return fp;
}

/**
 *
 * Close configuration file. In case a temporary file has been
 * create the file shall be automatically removed.
 *
 */

void mgtn_close_configuration_file(
  FILE *fp)
{
  int i;
  
  i = fclose(fp);

  if (i != 0)
    error("Could not properly close configuration file.\n");
}
    

/**
 *
 * Generate temporal configuration file with default values
 * overwritten with parameters defined through argument.
 *
 */

void mgtn_generate_configuration_file(
  FILE *fp,
  int model,
  int scheme,
  double alpha,
  int vcycles)
{
  int i;

  printMsg("Generating temporal configuration file with command-line specified arguments.\n");

  fprintf(fp, "# Scheme and resolution levels\n");
  fprintf(fp, "%d 6\n", scheme);

  fprintf(fp, "# Functional: data term, regularization term\n");
  switch (model) {
    case MODEL_LINEAR_QUADRATIC: 
      fprintf(fp, "%d %d\n", FUNCTIONAL_DATA_LINEAR, FUNCTIONAL_REGUL_QUADRATIC); 
      break;
    case MODEL_NON_LINEAR_QUADRATIC: 
      fprintf(fp, "%d %d\n", FUNCTIONAL_DATA_NON_LINEAR, FUNCTIONAL_REGUL_QUADRATIC); 
      break;
    case MODEL_LINEAR_TV:  
      fprintf(fp, "%d %d\n", FUNCTIONAL_DATA_LINEAR, FUNCTIONAL_REGUL_TV); 
      break;
    case MODEL_NON_LINEAR_TV:  
      fprintf(fp, "%d %d\n", FUNCTIONAL_DATA_NON_LINEAR, FUNCTIONAL_REGUL_TV); 
      break;
    default: error("No valid model value.\n");
  }

  fprintf(fp, "# Functional parameters: alpha, mu, outlier\n");
  fprintf(fp, "%f 0.05 15\n", alpha);

  fprintf(fp, "# Prefiltering: length, standard deviation\n");
  switch (model) {
    case MODEL_LINEAR_QUADRATIC:
    case MODEL_LINEAR_TV:
      fprintf(fp, "7 1.0\n");
      break;

    case MODEL_NON_LINEAR_QUADRATIC:
    case MODEL_NON_LINEAR_TV:
      fprintf(fp, "7 0.7\n");
      break;
  }

  fprintf(fp, "# Image gradient: type, length, number of interpolated pixels\n");
  switch (model) {
    case MODEL_LINEAR_QUADRATIC:
    case MODEL_LINEAR_TV:
      fprintf(fp, "3 1 1\n");
      break;

    case MODEL_NON_LINEAR_QUADRATIC:
    case MODEL_NON_LINEAR_TV:
      fprintf(fp, "1 5 1\n");
      break;
  }

  fprintf(fp, "# Truncated Newton parameters: eps_f, eps_w, eps_g, max_inner\n");
  fprintf(fp, "1e-5 1e-5 1e-5 20\n");
  fprintf(fp, "# Maxouter in multiresolution\n");
  fprintf(fp, "15 25\n");
  fprintf(fp, "# Multigrid: v or w-cycle, check descent, beta, kappa_g, eps_{Rg}\n");
  fprintf(fp, "1 1 10 0.5 0.5\n");
  fprintf(fp, "# Number of cycles at each resolution level\n");
  for(i = 0; i < 6; i++) fprintf(fp, "%d ", vcycles); fprintf(fp, "\n");
  fprintf(fp, "# Pre-optimize iterations\n");
  fprintf(fp, "1 1 1 1 1 1\n");
  fprintf(fp, "# Post-optimize iterations\n");
  fprintf(fp, "1 1 1 1 1 1\n");

  /* Reset file position to beginning */
  rewind(fp);
}

/**
 *
 *  Read configuration file 
 *
 */

void mgtn_read_configuration_file(
    FILE *fp,
    struct resources *DD)
{ 
  char str[200];
  char line[200], *p;

  int i, j, flag_has_multigrid;

  printMsg("These are the contents of the configuration file.\n");

  /* Get a comment */

  printMsg("------------------------------------------------\n");
  xfgets(line, 200, fp);

  /* Get data */

  xfgets(line, 200, fp);

  p = strtok(line, " \n");
  if (!p) 
    error("Missing scheme to be used.\n"); 

  DD->param.scheme = atoi(p);
  printMsg("Scheme to be used: %d ", DD->param.scheme);

  flag_has_multigrid = 0;  /* avoid complaining of compiler */

  switch (DD->param.scheme)
  {
    case SCHEME_MULTIRESOLUTION:
      printMsg("(multiresolution)\n");
      break;

    case SCHEME_FMG:
      flag_has_multigrid = 1;
      printMsg("(full multigrid)\n");
      break;

    default:
      printMsg("\n");
      error("No valid value for the scheme.\n");
  }

  p = strtok(NULL, " \n");
  if (!p) 
    error("Missing number of resolution levels.\n");

  j = atoi(p);
  printMsg("Resolution levels: %d\n", j);

  if (j < 1) 
  {
    sprintf(str, "The number of resolution levels should be greater or equal than %d\n", MIN_RESOLUTION_LEVELS);
    error(str);
  }

  if (j > MAX_RESOLUTION_LEVELS)
  {
    sprintf(str, "The number of resolution levels should be less or equal than %d\n", MAX_RESOLUTION_LEVELS);
    error(str);
  }

  DD->param.max_level = --j;

  /* Get a comment */

  printMsg("------------------------------------------------\n");
  xfgets(line, 200, fp);

  /* Get data associated to the general aspects of the functional */

  xfgets(line, 200, fp);

  p = strtok(line, " \n");
  if (!p) 
    error("Missing type of data term.\n");

  DD->param.functional_data_type = atoi(p);
  printMsg("Data term type: %d ", DD->param.functional_data_type);

  switch (DD->param.functional_data_type)
  {
    case FUNCTIONAL_DATA_NON_LINEAR:
      printMsg("(non-linear based)\n");
      break;
    case FUNCTIONAL_DATA_LINEAR:
      printMsg("(linear based)\n");
      break;
    default:
      printMsg("\n");
      error("No valid value for data term type.\n");
  }

  p = strtok(NULL, " \n");
  if (!p) 
    error("Missing regularitzation term.\n");

  DD->param.functional_regul_type = atoi(p);
  printMsg("Regularizer type: %d ", DD->param.functional_regul_type);

  switch (DD->param.functional_regul_type)
  {
    case FUNCTIONAL_REGUL_QUADRATIC:
      printMsg("(quadratic)\n");
      break;
    case FUNCTIONAL_REGUL_TV:
      printMsg("(total variation)\n");
      break;
    default:
      printMsg("\n");
      error("No valid regularizer type.\n");
  }

  /* Get a comment */

  printMsg("------------------------------------------------\n");
  xfgets(line, 200, fp);

  /* Get data */

  xfgets(line, 200, fp);

  p = strtok(line, " \n");
  if (!p) 
    error("Missing alpha (weight of regularizer).\n");

  DD->param.alpha = atof(p);
  printMsg("Alpha (weight of regularizer): %7.5f\n", DD->param.alpha);

  if (DD->param.alpha < MIN_ALPHA)
  {
    sprintf(str, "The value of alpha should be greater or equal than %f\n", MIN_ALPHA);
    error(str);
  }

  if (DD->param.alpha > MAX_ALPHA)
  {
    sprintf(str, "The value of alpha should be less or equal than %f\n", MAX_ALPHA);
    error(str);
  }

  p = strtok(NULL, " \n");
  if (!p) 
    error("Missing mu (for TV regularizer).\n");

  DD->param.mu = atof(p);
  printMsg("Mu value (for TV regularizer): %7.5f\n", DD->param.mu);

  if (DD->param.mu <= 0.0)
    error("The value of mu should be greater than 0.0\n");

  p = strtok(NULL, " \n");
  if (!p) 
    error("Missing value of outlier.\n");

  DD->param.outlier = atof(p);
  printMsg("Outlier value (for robust energy): %7.5f\n", DD->param.outlier);

  if (DD->param.outlier <= 0.0)
    error("The value of outlier should be greater than 0.0\n");

  /* Get comment */

  printMsg("------------------------------------------------\n");
  xfgets(line, 200, fp);

  /* Process prefilter information */

  xfgets(line, 200, fp);

  p = strtok(line, " \n");
  if (!p) 
    error("Missing filter length (for the prefiltering).\n");

  DD->param.prefilter_size = atoi(p);

  printMsg("Length of the prefilter: %d\n", DD->param.prefilter_size);

  if (DD->param.prefilter_size <= 1)
    error("The value of length should be greater or equal than 1\n");

  if (!ODD(DD->param.prefilter_size)) 
    error("Filter length (for the prefiltering) should be odd.\n");

  p = strtok(NULL, " \n");
  if (!p) 
    error("Missing sigma value (for the prefiltering).\n");

  DD->param.prefilter_sigma  = atof(p);
  printMsg("Prefilter sigma value: %7.5f\n", DD->param.prefilter_sigma);

  if (DD->param.prefilter_sigma <= 0.0)
    error("The value of prefilter sigma should be greater than 0.0\n");

  /* Get comment */

  printMsg("------------------------------------------------\n");
  xfgets(line, 200, fp);

  /* Get data for image gradient computation */

  xfgets(line, 200, fp);

  p = strtok(line, " \n");
  if (!p) 
    error("Missing filter type (for image gradient computation).\n");

  DD->param.gradient_filter_type = atoi(p);

  if ((DD->param.functional_data_type == FUNCTIONAL_DATA_LINEAR) &&
      (DD->param.gradient_filter_type != FILTER_OFE_HS)) {
    printMsg("Setting Horn & Schunk filter since linear OFE is used.\n");
    DD->param.gradient_filter_type = FILTER_OFE_HS;
  }

  printMsg("Gradient filter type: %d ", DD->param.gradient_filter_type);

  switch (DD->param.gradient_filter_type)
  {
    case FILTER_SIMONCELLI:
      printMsg("(Simoncelli's stencil)\n");
      break;
    case FILTER_GAUSS:
      printMsg("(Gaussian's stencil)\n");
      break;
    case FILTER_OFE_HS:
      printMsg("(Horn & Schunk stencil)\n");
      break;
    default:
      printMsg("\n");
      error("No valid filter (for image gradient computation).\n");
  }

  p = strtok(NULL, " \n");
  if (!p) 
    error("Missing stencil size (for image gradient computation).\n");

  if (DD->param.gradient_filter_type != FILTER_OFE_HS)
  {
    DD->param.gradient_filter_size = atoi(p);
    printMsg("Gradient filter size: %d\n", DD->param.gradient_filter_size);

    if (DD->param.gradient_filter_size < 3) 
      error("Stencil size (for image gradient computation) is too small, set at least to 5.\n");
  }
  else
    printMsg("Gradient filter size: ignored.\n");

  p = strtok(NULL, " \n");
  if (!p) 
    error("Missing number of interpolated samples (for image gradient computation).\n");

  if (DD->param.functional_data_type == FUNCTIONAL_DATA_LINEAR) {
    printMsg("Setting interpolation to 1 since OFE is used.\n");
    DD->param.interpolated_samples = 1;
  }
  else
    DD->param.interpolated_samples = atoi(p);

  printMsg("Number of interpolated samples: %d\n", DD->param.interpolated_samples);

  if (DD->param.interpolated_samples < 1)
    error("Interpolated samples should be an odd number greater or equal than 1.\n");

  if ((DD->param.interpolated_samples != 1) && (ODD(DD->param.interpolated_samples))) 
    error("Interpolated samples should be either equal o 1 or an even number.\n");

  /* Get a comment */

  printMsg("------------------------------------------------\n");
  xfgets(line, 200, fp);

  /* Get data for truncated newton parameters */

  xfgets(line, 200, fp);

  p = strtok(line, " \n");
  if (!p) 
    error("Missing ftol value.\n");

  DD->param.ftol = atof(p);
  printMsg("ftol: %e ", DD->param.ftol);

  if (DD->param.ftol < 0.0) 
    printMsg("(set default value by TNC)\n");
  else
    printMsg("\n");

  p = strtok(NULL, " \n");
  if (!p) 
    error("Missing xtol value.\n");

  DD->param.xtol = atof(p);
  printMsg("xtol: %e ", DD->param.xtol);

  if (DD->param.xtol < 0.0) 
    printMsg("(set default value by TNC)\n");
  else
    printMsg("\n");

  p = strtok(NULL, " \n");
  if (!p) 
    error("Missing pgtol value.\n");

  DD->param.pgtol = atof(p);
  printMsg("pgtol: %e ", DD->param.pgtol);

  if (DD->param.pgtol < 0.0) 
    printMsg(" (set default value by TNC)\n");
  else
    printMsg("\n");

  p = strtok(NULL, " \n");
  if (!p) 
    error("Missing internal iterations value.\n");

  DD->param.maxit_tn_internal = atoi(p);
  printMsg("Maxit internal iterations (maxinner): %d\n", DD->param.maxit_tn_internal);

  if (DD->param.maxit_tn_internal <= 0) 
    error("Number of internal iterations should be greater than 0.\n");

  /* Get comment */

  printMsg("------------------------------------------------\n");
  xfgets(line, 200, fp);

  /* Get data */

  xfgets(line, 200, fp);

  p = strtok(line, " \n");
  if (!p) 
    error("Missing number of iterations (for multiresolution).\n");

  DD->param.maxit_multiresolution = atoi(p);

  j = 0;

  if (DD->param.scheme == SCHEME_MULTIRESOLUTION) {
    printMsg("Maxit of TN in multiresolution (maxouter): %d\n", DD->param.maxit_multiresolution);
    j = DD->param.maxit_multiresolution;

    if (DD->param.max_level == 0) {
      if (j < MIN_UNILEVEL_ITERATIONS) {
	sprintf(str, "Unilevel: max-iterations should be greater or equal than %d\n", MIN_UNILEVEL_ITERATIONS);
	error(str);
      }

      if (j > MAX_UNILEVEL_ITERATIONS) {
	sprintf(str, "Unilevel: max-iterations should be less or equal than %d\n", MAX_UNILEVEL_ITERATIONS);
	error(str);
      }
    } else {
      if (j < MIN_MR_ITERATIONS) {
	sprintf(str, "Multiresolution: max-iterations should be greater or equal than %d\n", MIN_MR_ITERATIONS);
	error(str);
      }

      if (j > MAX_MR_ITERATIONS) {
	sprintf(str, "Multiresolution: max-iterations should be less or equal than %d\n", MAX_MR_ITERATIONS);
	error(str);
      }
    }
  } else {
    printMsg("Maxit of TN in multiresolution: skipped\n");

    p = strtok(NULL, " \n");
    if (!p) 
      error("Missing maximum number of iterations (for multigrid's coarsest level).\n");

    DD->param.maxit_coarse_multigrid = atoi(p);
    printMsg("Maxit in multigrid's coarsest level: %d\n", DD->param.maxit_coarse_multigrid);

    j = DD->param.maxit_coarse_multigrid;

    if (j < MIN_MR_ITERATIONS) {
      sprintf(str, "Multigrid: max-iterations should be greater or equal than %d\n", MIN_MR_ITERATIONS);
      error(str);
    }

    if (j > MAX_MR_ITERATIONS) {
      sprintf(str, "Multigrid: max-iterations should be less or equal than %d\n", MAX_MR_ITERATIONS);
      error(str);
    }
  }

  /* These parameters are related to multigrid */

  printMsg("------------------------------------------------\n");

  if (!flag_has_multigrid)
    printMsg("Multigrid options discarded\n");
  else {  

    /* Get comment */

    xfgets(line, 200, fp);

    /* Get data */

    xfgets(line, 200, fp);

    p = strtok(line, " \n");
    if (!p) 
      error("Missing type of multigrid v-cycle.\n");

    DD->param.type_multigrid = atoi(p);
    printMsg("Multigrid v-cycle type: %d ", DD->param.type_multigrid);

    switch (DD->param.type_multigrid)
    {
      case V_CYCLE:
	printMsg(" (V-cycle)\n");
	break;
      case W_CYCLE:
	printMsg(" (W-cycle)\n");
	break;
      default:
	printMsg("\n");
	error("No valid value for multigrid v-cycle type.\n");
    }

    p = strtok(NULL, " \n");
    if (!p) 
      error("Missing value for line minimization behaviour.\n");

    DD->param.check_descent = atoi(p);
    printMsg("Check if correction step already provides descent: %d ", DD->param.check_descent);

    switch (DD->param.check_descent)
    {
      case MG_CHECK_DESCENT_NO:
	printMsg("(deactivated)\n");
	break;

      case MG_CHECK_DESCENT_YES:
	printMsg("(activated)\n");
	break;

      default:
	printMsg("\n");
	error("No valid value for check descent.\n");
    }

    p = strtok(NULL, " \n");
    if (!p) 
      error("Missing gamma value for constraint in postsmooth.\n");

    DD->param.beta_constraint = atof(p);
    printMsg("Beta value for constraint: %E ", DD->param.beta_constraint);

    if (DD->param.beta_constraint < 0.0)
      printMsg("(no constraint)\n");
    else
      printMsg("\n");

    p = strtok(NULL, " \n");
    if (!p) 
      error("missing 'a' presmoothing threshold.\n");

    DD->param.a_presmooth_cond = atof(p);
    printMsg("Presmoothing value 'a': %E ", DD->param.a_presmooth_cond);

    if (DD->param.a_presmooth_cond < 0.0)
      printMsg("(deactivated)\n");
    else
      printMsg("\n");

    p = strtok(NULL, " \n");
    if (!p) 
      error("missing 'e' presmoothing threshold.\n");

    DD->param.e_presmooth_cond = atof(p);
    printMsg("Presmoothing value 'e': %E ", DD->param.e_presmooth_cond);

    if (DD->param.e_presmooth_cond < 0.0)
      printMsg("(deactivated)\n");
    else
      printMsg("\n");

    if (DD->param.a_presmooth_cond * DD->param.e_presmooth_cond <= 0.0) 
      error("The latter two values cannot have different sign.\n");

    /* Get comment */

    xfgets(line, 200, fp);

    /* Get data */

    xfgets(line, 200, fp);

    DD->param.maxit_v_cycle = (int *) xmalloc(sizeof(int) * (DD->param.max_level + 1));

    printMsg("V-cycle iterations: ");

    p = strtok(line, " \n");

    for(i = 0; i <= DD->param.max_level; i++)
    {
      if (!p) {
	printMsg("\n");
	error("Missing data in the iterations.\n"); 
      }

      j = atoi(p);
      printMsg("%d, ", j);
      DD->param.maxit_v_cycle[i] = j;

      if (j < MIN_V_CYCLES)
      {	
	sprintf(str, "V-cycle iterations should be greater or equal than %d.\n", MIN_V_CYCLES);
	error(str);
      }

      if (j > MAX_V_CYCLES)
      {
	sprintf(str, "V-cycle iterations should be less or equal than %d.\n", MAX_V_CYCLES);
	error(str);
      }

      p = strtok(NULL, " \n");
    }  

    printMsg("\n");

    /* Get comment */

    xfgets(line, 200, fp);

    /* Get data */

    xfgets(line, 200, fp);

    DD->param.maxit_presmoothing = (int *) xmalloc(sizeof(int) * (DD->param.max_level + 1));

    printMsg("Max presmoothing iterations: ");

    p = strtok(line, " \n");

    for(i = 0; i <= DD->param.max_level; i++)
    {
      if (!p) {
	printMsg("\n");
	error("Missing data in the presmoothing iterations.\n"); 
      }

      j = atoi(p);
      printMsg("%d, ", j);
      DD->param.maxit_presmoothing[i] = j;

      if (j < 0)
	error("Presmoothing iterations should be positive (including zero).\n");

      p = strtok(NULL, " \n");
    }  

    printMsg("\n");

    /* Get comment */

    xfgets(line, 200, fp);

    /* Get data */

    xfgets(line, 200, fp);

    DD->param.maxit_postsmoothing = (int *) xmalloc(sizeof(int) * (DD->param.max_level + 1));

    printMsg("Max postsmoothing iterations: ");

    p = strtok(line, " \n");

    for(i = 0; i <= DD->param.max_level; i++)
    {
      if (!p) {
	printMsg("\n");
	error("Missing data in the postsmoothing iterations.\n"); 
      }

      j = atoi(p);
      printMsg("%d, ", j);
      DD->param.maxit_postsmoothing[i] = j;

      if (j < 0)
	error("Postsmoothing iterations should be positive (including zero)\n");

      p = strtok(NULL, " \n");
    }  

    printMsg("\n");
  }

  printMsg("------------------------------------------------\n");
}

/**
 *
 * Read ground truth flow from disc if available. Flow is
 * assumed to be in flo format. 
 *
 */

void mgtn_groundtruth_read(
    char *fname,
    struct resources *DD)
{
  double *p_u,*p_v;
  float header, u, v;

  FILE *fp;

  struct vector2 coord_pix;
  int nrow, ncol;

  fp = fopen(fname, "r");

  if (!fp) 
    error("Could not open ground truth file.\n");

  printMsg("Reading ground truth flow '%s'\n", fname);

  xfread(&header, sizeof(float), 1, fp);

  if (header != 202021.25) {
    printMsg("File does not contain a valid header. Ignoring contents\n");
    fclose(fp);
    return;
  }

  DD->curr_level = 0;
  DD->have_groundtruth_flow = 1;

  xfread(&ncol, sizeof(float), 1, fp);
  xfread(&nrow, sizeof(float), 1, fp);

  if ((nrow != DD->ori_in0->nrow) || (ncol != DD->ori_in0->ncol)) {
    printMsg("Ground truth size is not coincident with image size. Ignoring contents.\n");
    fclose(fp);
    return;
  }

  DD->groundtruth_flow_u = mgtn_image_alloc(NULL, nrow, ncol);
  DD->groundtruth_flow_v = mgtn_image_alloc(NULL, nrow, ncol);

  p_u = DD->groundtruth_flow_u->gray;
  p_v = DD->groundtruth_flow_v->gray;

  for(coord_pix.y = 0; coord_pix.y < nrow; coord_pix.y++)
    for(coord_pix.x = 0; coord_pix.x < ncol; coord_pix.x++, p_u++, p_v++)
    {
      xfread(&u, sizeof(float), 1, fp);
      xfread(&v, sizeof(float), 1, fp);

      *p_u = u;
      *p_v = v;
    }

  fclose(fp);
}

/**
 *
 * Write flow in flo format. 
 *
 */

void mgtn_flow_write(
    char *fname,
    struct resources *DD)
{
  union { char s[4]; float f; } pieh = {"PIEH"};

  FILE *fp;
  struct image *flow_u, *flow_v;

  int ncol, nrow, i, j;
  double *p_u, *p_v;
  float u, v;

  if (pieh.f != 202021.25) {
    error("The code for writing flows works for little-endian architectures.\n");
    error("You should adapt the code to your architectures. Sorry!\n");
    return;
  }

  flow_u = GET_CURRENT_IMAGE(DD, u);
  flow_v = GET_CURRENT_IMAGE(DD, v);

  ncol = flow_u->ncol; 
  nrow = flow_u->nrow;

  p_u = flow_u->gray; 
  p_v = flow_v->gray; 

  fp = fopen(fname, "w");

  if (!fp) 
    error("Could not open output flow for writing\n");

  printMsg("Writing flow in file '%s'\n", fname);

  xfwrite(&pieh.f, sizeof(float), 1, fp);
  xfwrite(&ncol, sizeof(int), 1, fp);
  xfwrite(&nrow, sizeof(int), 1, fp);

  for(i = 0; i < nrow; i++)
  {
    for(j = 0; j < ncol; j++, p_u++, p_v++)
    {
      u = *p_u;
      v = *p_v;

      xfwrite(&u, sizeof(float), 1, fp);
      xfwrite(&v, sizeof(float), 1, fp);
    }
  }

  fclose(fp);
}  

/**
 * 
 *  Write backwards compensated image
 *
 */ 

void mgtn_backcomp_write(
    char *fname,
    struct resources *DD)
{
  struct image *u, *v, *backcomp;
  struct vector2 pix;

  double value;
  int i, j, nrow, ncol, max_x, max_y, offset;

  u = GET_CURRENT_IMAGE(DD, u);
  v = GET_CURRENT_IMAGE(DD, v);

  ncol = u->ncol; 
  nrow = u->nrow;

  max_x = ncol - 1;
  max_y = nrow - 1;

  backcomp = mgtn_image_alloc(NULL, nrow, ncol); 

  /* Proceed through all the image and perform
   * backwards compensation */

  for(i = 0; i < nrow; i++)
  {
    for(j = 0; j < ncol; j++)
    {
      offset = i * ncol + j;

      pix.x = (double) j;
      pix.y = (double) i;

      pix.x += u->gray[offset];
      pix.y += v->gray[offset];

      if ((pix.x >= 0) && (pix.x <= max_x) && 
	  (pix.y >= 0) && (pix.y <= max_y))
      {
	value = mgtn_image_interpolate(1, &pix, DD); 

	if (value > 255.0)
	  backcomp->gray[offset] = 255.0;
	else if (value < 0.0)
	  backcomp->gray[offset] = 0.0;
	else
	  backcomp->gray[offset] = value;
      }
      else
	backcomp->gray[offset] = 0;
    }
  }

  printMsg("Writing backwards compensated image in '%s'\n", fname);
  mgtn_pgm_write_image(backcomp, fname); 

  mgtn_image_delete(backcomp); 
}

