//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Affero General Public License as
//  published by the Free Software Foundation, either version 3 of the
//  License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU Affero General Public License for more details.
//
//  You should have received a copy of the GNU Affero General Public License
//  along with this program. If not, see <http://www.gnu.org/licenses/>.

// Copyright (c) Lluis Garrido <lluis.garrido@ub.edu> and <lgoster@gmail.com>

#include "mgtn.h"

/**
 *
 * Creates a new empty vector structure 
 *
 */

struct vector *mgtn_vector_new()
{
  struct vector *vector;

  vector = (struct vector *) xmalloc(sizeof(struct vector));

  vector->size = 0;
  vector->values = NULL;

  return (vector);
}

/** 
 *
 * Allocates a vector structure with N positions
 *
 */

struct vector *mgtn_vector_alloc(
     struct vector *vector,
     int N)
{
  int size;
  struct vector *tmp;

  if (vector == NULL)
    tmp = mgtn_vector_new();
  else
  {
    tmp = vector;

    if (tmp->values != NULL)
      error("Attempting to alloc a vector which is already allocated.\n");
  }

  size = N * sizeof(double);
  if (size <= 0)
    error("Cannot allocate a vector with zero or negative size.\n");

  tmp->size = N;
  tmp->values = (double *) xmalloc(size);

  return tmp;
}

/**
 *
 * Deallocates a vector 
 *
 */

void mgtn_vector_delete(
  struct vector *vector)
{
  if (vector == NULL)
    error("Cannot delete vector: structure is NULL\n");

  if (vector->values != NULL) 
  {
    free(vector->values);
    vector->values = NULL;
  }

  free(vector);
}

/** 
 *
 * Clear the array of a fsignal with value
 *
 */

void mgtn_vector_clear(
  struct vector *vector,
  double value)
{
  double *p;
  int i;

  if ((!vector) || (!vector->values)) 
    error("Vector is not allocated.\n");

  p = vector->values;
  for (i = 0; i < vector->size; i++, p++)
    *p = value;
}






